/*
Derived from Examples in 6.2, Network Security with OpenSSL by Pravir Chandra, Matt Messier, John Viega. 
Copyright  2002 O'Reilly & Associates, Inc.

This duplicates the functionality of:
openssl enc -bf-cbc

The format of the data is:
Salted__
salt[8]
data in blocks of 8.
It's very easy.

The default openssl enc -bf-cbc uses md5 hash to generate the key and iv. This 
is bad because md5 has a collision problem. Collisions can be generated 
*/

#include <openssl/evp.h>
#include <openssl/rand.h>
#include <stdio.h>
#include <string.h>

// use a 256-bit key like they say instead of a 128-bit key.
#define KEY_LENGTH EVP_MAX_KEY_LENGTH
// Header means: blowfish cbc, iv is the first 8 bytes, 
// salt is the next 8 bytes, and the password is sha1 hashed.
#define BF_HEADER "Salted__"

#define MAX_PACKET_SIZE 1024

#ifdef _DEBUG
#define VERBOSE 1
#else
#define VERBOSE 0
#endif

// The idea is that I have 2 programs and one library in one .c file
#ifdef ENCRYPT
#define encrypt_main main
#endif
#ifdef DECRYPT
#define decrypt_main main
#endif

#ifdef _OPENBSD
#define RANDOM_DEVICE "/dev/srandom"
#else
#define RANDOM_DEVICE "/dev/random"
#endif

#include <ssl_bf2a.h>

void select_random_key(BYTE *key, int b)
{
    RAND_bytes(key, b);
}

void print_hex(BYTE *data, int b)
{
    int i;

    for (i = 0;  i < b - 1;  i++)
        fprintf(stderr, "%02hhX:", data[i]);
    fprintf(stderr, "%02hhX\n", data[b - 1]);
}

unsigned char *simple_digest(char *buf, unsigned int len, int *olen)
{
    const EVP_MD  *m;
    EVP_MD_CTX    ctx;
    unsigned char *ret;

    char alg[] = "sha1";

    OpenSSL_add_all_digests(  );
    if (!(m = EVP_get_digestbyname(alg)))
        return NULL;
    if (!(ret = (unsigned char *)malloc(EVP_MAX_MD_SIZE)))
        return NULL;
    EVP_DigestInit(&ctx, m);
    EVP_DigestUpdate(&ctx, buf, len);
    EVP_DigestFinal(&ctx, ret, olen);
    return ret;
}

void get_password(char *pass, int len, int encrypting)
{
    for (;;)
    {
        // If encrypting == 1, verify password.
        pass[0]='\0';
        int i = EVP_read_pw_string((char *)pass, len, "Password:", encrypting);
        if (i == 0)
        {
            if (pass[0] == '\0')
            {
                return;
            }
            //str = pass;
            break;
        }
        if (i < 0)
        {
            fprintf(stderr,"Error: bad password read\n");
            return;
        }
    }
}

void generate_key(BYTE *key, const char *password, const BYTE *salt, BYTE *iv, int b)
{
    const EVP_CIPHER *cipher = NULL;
    cipher = EVP_bf_cbc(); // Error: EVP_get_cipherbyname("bf-cbc");
    const EVP_MD *dgst = NULL;
    dgst = EVP_md5();
    // This is a pretty interesting system.
    // read openssl-xxx/doc/EVP_BytesToKey.pod for interesting info.
    EVP_BytesToKey(cipher, dgst, salt, (BYTE *)password, 
        strlen(password), 1, key, iv);
}

void select_random_salt(char *iv, int b)
{
    RAND_pseudo_bytes(iv, b);
}

int seed_prng(int bytes)
{
    // This slows you down if you use it very much.
    if (!RAND_load_file(RANDOM_DEVICE, bytes))
        return 0;
    return 1;
}

char *encrypt_data(EVP_CIPHER_CTX *ctx, char *data, int inl, int *rb)
{
    char *ret;
    int  i, tmp, ol;

    ol = 0;
    // Note: you must free this memory.
    ret = (char *)malloc(inl + EVP_CIPHER_CTX_block_size(ctx));
    // 104 is divisible by 8 which is important.
    for (i = 0;  i < inl / 104;  i++)
    {
        EVP_EncryptUpdate(ctx, &ret[ol], &tmp, &data[ol], 104);
        ol += tmp;
    }
    if (inl % 104)
    {
        /*char last[104];
        // FIXME: I don't know why I need to add 4, but I do.
        // This is likely to cause errors in the future.
        memcpy(last, &data[ol], inl % 104);
        last[inl%104] = 0;
        fprintf(stderr, "last encrypt: '%s'\n", last);
        */
        EVP_EncryptUpdate(ctx, &ret[ol], &tmp, &data[ol], inl % 104);
        ol += tmp;
    }
    EVP_EncryptFinal(ctx, &ret[ol], &tmp);
    *rb = ol + tmp;
    return ret;
}

char *decrypt_data(EVP_CIPHER_CTX *ctx, char *ct, int inl)
{
    // Note: you must free this memory.
    char *pt = (char *)malloc(inl + EVP_CIPHER_CTX_block_size(ctx));
    int  ol;

    EVP_DecryptUpdate(ctx, pt, &ol, ct, inl);
    if (!ol) /* there's no block to decrypt */
    {
        free(pt);
        return NULL;
    }
    return pt;
}


/*
The encrypt function:

needs to be called many times in a row.
give it data and a password and you get a packet of data.

The decrypt function:

needs to be called many times in a row.
give it a packet and a password and you get data.

Don't force the developer to do any salt boloney.
*/
int encrypt_packet(BYTE *data, int data_size, char *password, BYTE *out, int *enc_size)
{
    EVP_CIPHER_CTX ctx;
    BYTE           key[KEY_LENGTH]; 
    BYTE           iv[EVP_MAX_IV_LENGTH];
    BYTE           salt[PKCS5_SALT_LEN];

    // PKCS5_SALT_LEN will always be 8.
    select_random_salt(salt, PKCS5_SALT_LEN);
    
    generate_key(key, password, salt, iv, KEY_LENGTH);
    if(VERBOSE) print_hex(key, KEY_LENGTH);
    if(key == 0)
    {
        fprintf(stderr, "Error: Key generation failed.\n");
        return 1;
    }
    // iv is selected by the password+salt.

    EVP_EncryptInit(&ctx, EVP_bf_cbc(), key, iv);
    char *enc_data = encrypt_data(&ctx, data, data_size, enc_size);
    //printf("Ciphertext is %d bytes.\n", enc_size);
    memcpy(out, BF_HEADER, 8);
    memcpy(out+8, salt, PKCS5_SALT_LEN);
    memcpy(out+8+PKCS5_SALT_LEN, enc_data, *enc_size);
    // Add the header length and the salt length.
    *enc_size += 8 + PKCS5_SALT_LEN;
    // FIXME: Memory leak. *** glibc detected *** double free or corruption (!prev): 0x0804d1c0 ***
    //free(enc_data);
    return 0;
}

int decrypt_packet(BYTE *enc_data, int enc_size, char *password, BYTE *out, int *data_size)
{
    EVP_CIPHER_CTX ctx;
    BYTE           key[KEY_LENGTH]; 
    BYTE           iv[EVP_MAX_IV_LENGTH];
    BYTE           salt[PKCS5_SALT_LEN];

    char *dec_data;
    char *ptr_enc;
    char  final[EVP_MAX_BLOCK_LENGTH];
    int   ret = 0;

    // Take the encrypted data from the stdin.
    if(0 != strncmp(enc_data, BF_HEADER, 8))
    {
        fprintf(stderr, "Error: Header not found.\n");
        return 1;
    }
    memcpy(salt, enc_data + 8, PKCS5_SALT_LEN);
    generate_key(key, password, salt, iv, KEY_LENGTH);
    if(VERBOSE) print_hex(key, KEY_LENGTH);
    ptr_enc = enc_data + 16;
    enc_size = enc_size - 16;
    
    EVP_DecryptInit(&ctx, EVP_bf_cbc(), key, iv);
    int j, k = 0;
    for(j=0; j<enc_size; j+=8)
    {
        dec_data = decrypt_data(&ctx, ptr_enc + j, 8);
        // The first one is always null, ignore it or fix it.
        // TODO: You should probably try to understand it as well.
        if(dec_data != 0)
        {
            memcpy(out + k, dec_data, 8);
            free(dec_data);
            // Counter for output.
            k += 8;
        }
        //else
        //    fprintf(stderr, "Error: Decrypt failed, null data.\n");
    }
    if (!EVP_DecryptFinal(&ctx, final, data_size))
    {
        fprintf(stderr, "Error: Decrypt Final, padding incorrect.\n");
        ret = 1;
        //abort();
    }
    memcpy(out + k, final, *data_size);
    // Data size is set in DecryptFinal to the last size (0-7).
    // Add k (output counter) to get the correct total size.
    *data_size += k;
    return ret;
}

int encrypt_main(int argc, char *argv[])
{
    if (!seed_prng(4))
    {
        fprintf(stderr, "Error: Unable to seed the PRNG.\n");
        abort();
    }

    char           enc_data[MAX_PACKET_SIZE + 24];
    char           data[MAX_PACKET_SIZE];
    char           password[256];
    
    // Secure password retreival algorithm from enc
    get_password(password, 256, 1);
    
    // Read binary data up to 1024 bytes from the stdin.
    int data_size = fread(data, 1, MAX_PACKET_SIZE, stdin);
    //data[data_size] = 0;
    //fprintf(stderr, "data: '%i'\n", data_size);

    int enc_size = -1;
    int ret = encrypt_packet(data, data_size, password, enc_data, &enc_size);
    
    fwrite(enc_data, enc_size, 1, stdout);
    
    return ret;
}

int decrypt_main(int argc, char *argv[])
{
    if (!seed_prng(4))
    {
        fprintf(stderr, "Error: Unable to seed the PRNG.\n");
        abort();
    }

    char           enc_data[MAX_PACKET_SIZE + 24];
    char           data[MAX_PACKET_SIZE];
    char           password[256];
    
    // Secure password retreival algorithm from enc
    get_password(password, 256, 0);
    
    // Read binary data up to 1024 bytes from the stdin.
    int enc_size = fread(enc_data, 1, MAX_PACKET_SIZE + 24, stdin);
    int data_size = -1;
    int ret = decrypt_packet(enc_data, enc_size, password, data, &data_size);
    
    fwrite(data, data_size, 1, stdout);
    
    return ret;
}



